from typing import Dict, Tuple

class Pipeline:

    _dataFrameLoaders = dict()
    _transformations = dict()
    _sources = dict()

    def addDataFrameLoader(self, fun: callable):
        functionName = fun.__name__

        if functionName not in self._dataFrameLoaders:
            self._dataFrameLoaders[functionName] = fun

    def addTransformation(self, fun: callable, sources: Tuple[callable]):
        functionName = fun.__name__

        if functionName not in self._transformations:
            self._transformations[functionName] = fun
            self._sources[functionName] = sources

    def getSource(self, name) -> Tuple[callable]:
        return self._sources[name]

    def getTransformations(self) -> Dict[str, callable]:
        return self._transformations

    def getDataFrameLoaders(self) -> Dict[str, callable]:
        return self._dataFrameLoaders



import re
from injecta.container.ContainerInterface import ContainerInterface
from pyfonybundles.Bundle import Bundle
from databricksbundle.notebook.NotebookErrorHandler import setNotebookErrorHandler
from databricksbundle.detector import isDatabricks
from databricksbundle.notebook.helpers import getNotebookPath

class DatabricksBundle(Bundle):

    def boot(self, container: ContainerInterface):
        parameters = container.getParameters()

        if (
            isDatabricks()
            and parameters.databricksbundle.enableNotebookErrorHandler is True
            and not re.match('^/Users/', getNotebookPath())
        ):
            setNotebookErrorHandler()



from pyspark.sql import SparkSession
from databricksbundle.spark.SparkSessionLazy import SparkSessionLazy

class BasicSessionFactory:

    def create(self) -> SparkSessionLazy:
        def createLazy():
            return SparkSession \
                .builder \
                .getOrCreate()

        return SparkSessionLazy(createLazy)



from typing import Dict
from pyspark.sql.dataframe import DataFrame
from databricksbundle.pipeline.Pipeline import Pipeline

def invokePipeline(pipeline: Pipeline) -> Dict[str, DataFrame]:
    dataFramesByName = dict()

    for dataFrameLoaderName, dataFrameLoader in pipeline.getDataFrameLoaders().items():
        dataFramesByName[dataFrameLoaderName + '_df'] = dataFrameLoader()

    for transformationName, transformation in pipeline.getTransformations().items():
        def transformSource(source: callable):
            return dataFramesByName[source.__name__ + '_df']

        dataframesToUse = list(map(transformSource, pipeline.getSource(transformationName)))

        df = transformation(*dataframesToUse)
        dataFramesByName[transformationName + '_df'] = df

    return dataFramesByName



# pylint: disable = invalid-name
import sys
from typing import Tuple
from types import FunctionType
from databricksbundle.detector import isDatabricks
from databricksbundle.display import display

# ---------------------------------------------------------------------------------------------------------------------

def checkArgs(a):
    # called as @dataFrameLoader (without brackets and any arguments)
    if a and isinstance(a[0], FunctionType):
        raise Exception('Use @dataFrameLoader() instead of @dataFrameLoader please')

if isDatabricks():
    class dataFrameLoader:

        def __init__(self, *args, **kwargs):
            self._displayEnabled = kwargs.get('display', False)
            checkArgs(args)

        def __call__(self, fun, *args, **kwargs):
            g = fun.__globals__
            df = fun()
            g[fun.__name__ + '_df'] = df

            if self._displayEnabled:
                display(df)

            return fun

    class transformation:

        def __init__(self, *args: Tuple[callable], **kwargs):
            self._sources = args
            self._displayEnabled = kwargs.get('display', False)

        def __call__(self, fun, *args, **kwargs):
            g = fun.__globals__

            def transformSource(source: callable):
                return g[source.__name__ + '_df']

            dataframesToUse = list(map(transformSource, self._sources))

            df = fun(*dataframesToUse)
            g[fun.__name__ + '_df'] = df

            if self._displayEnabled:
                display(df)

            return fun

    class dataFrameSaver:

        def __init__(self, *args: Tuple[callable]):
            self._sources = args

        def __call__(self, fun, *args, **kwargs):
            g = fun.__globals__

            def transformSource(source: callable):
                return g[source.__name__ + '_df']

            dataframesToUse = list(map(transformSource, self._sources))

            fun(*dataframesToUse)

            return fun

# ---------------------------------------------------------------------------------------------------------------------

elif 'unittest' in sys.modules:
    class dataFrameLoader:

        def __init__(self, *args, **kwargs): # pylint: disable = unused-argument
            checkArgs(args)

        def __call__(self, fun, *args, **kwargs):
            return fun

    class transformation:

        def __init__(self, *args, **kwargs):
            pass

        def __call__(self, fun, *args, **kwargs):
            return fun

    class dataFrameSaver:

        def __init__(self, *args):
            pass

        def __call__(self, fun, *args, **kwargs):
            return fun

# ---------------------------------------------------------------------------------------------------------------------

else:
    from databricksbundle.pipeline.pipelineInvoker import invokePipeline
    from databricksbundle.pipeline.Pipeline import Pipeline

    pipeline = Pipeline()

    class dataFrameLoader:

        def __init__(self, *args, **kwargs): # pylint: disable = unused-argument
            checkArgs(args)

        def __call__(self, fun, *args, **kwargs):
            pipeline.addDataFrameLoader(fun)
            return fun

    class transformation:

        def __init__(self, *args: Tuple[callable], **kwargs): # pylint: disable = unused-argument
            self._sources = args

        def __call__(self, fun, *args, **kwargs):
            pipeline.addTransformation(fun, self._sources)
            return fun

    class dataFrameSaver:

        def __init__(self, *args: Tuple[callable]):
            self._sources = args

        def __call__(self, fun, *args, **kwargs):
            dataFramesByName = invokePipeline(pipeline)

            def transformSource(source: callable):
                return dataFramesByName[source.__name__ + '_df']

            dataframesToUse = list(map(transformSource, self._sources))

            fun(*dataframesToUse)

            return fun



from databricksbundle.dbutils.DatabricksConnectDbUtilsFactory import DatabricksConnectDbUtilsFactory
from databricksbundle.dbutils.DatabricksDbUtilsFactory import DatabricksDbUtilsFactory
from databricksbundle.dbutils.DbUtilsWrapper import DbUtilsWrapper
from databricksbundle.detector import isDatabricks

class AutodetectionDbUtilsFactory:

    def __init__(
        self,
        databricksConnectDbUtilsFactory: DatabricksConnectDbUtilsFactory,
        databricksDbUtilsFactory: DatabricksDbUtilsFactory,
    ):
        self.__databricksConnectDbUtilsFactory = databricksConnectDbUtilsFactory
        self.__databricksDbUtilsFactory = databricksDbUtilsFactory

    def create(self) -> DbUtilsWrapper:
        if isDatabricks():
            return self.__databricksDbUtilsFactory.create()

        return self.__databricksConnectDbUtilsFactory.create()



from databricksbundle.spark.DatabricksConnectSessionFactory import DatabricksConnectSessionFactory
from databricksbundle.spark.DatabricksSessionFactory import DatabricksSessionFactory
from databricksbundle.detector import isDatabricks
from databricksbundle.spark.SparkSessionLazy import SparkSessionLazy

class AutodetectionSessionFactory:

    def __init__(
        self,
        databricksConnectSessionFactory: DatabricksConnectSessionFactory,
        databricksSessionFactory: DatabricksSessionFactory,
    ):
        self.__databricksConnectSessionFactory = databricksConnectSessionFactory
        self.__databricksSessionFactory = databricksSessionFactory


    def create(self) -> SparkSessionLazy:
        if isDatabricks():
            return self.__databricksSessionFactory.create()

        return self.__databricksConnectSessionFactory.create()
